@@ -14,27 +14,20 @@ import shutil
 from logilab.mtconverter import need_guess, guess_mimetype_and_encoding
 
 from cubicweb import QueryError, ValidationError, Binary
-from cubicweb.selectors import objectify_selector
-from cubicweb.server import hook
+from cubicweb.server import hook, session
 from cubicweb.selectors import is_instance, score_entity, configuration_values
 
 from cubes.vcsfile import IMMUTABLE_ATTRIBUTES, bridge
 
-# XXX in cw >= 3.12.4
-@objectify_selector
-def is_user_session(cls, req, **kwargs):
-    return not req.is_internal_session
-
 def repo_cache_dir(config):
     directory = config['local-repo-cache-root']
     if not osp.isabs(directory):
         directory = osp.join(config.appdatahome, directory)
-        if not config.repairing: # don't change value during migration
-            config['local-repo-cache-root'] = directory
+    config.local_repo_cache_root = directory
     if not osp.exists(directory):
         try:
             os.makedirs(directory)
-        except:
+        except Exception:
             config.critical('could not find local repo cache directory; check '
                             'local-repo-cache-root option whose present value is %r)',
                             config['local-repo-cache-root'])
@@ -96,35 +89,32 @@ def missing_attribute_error(entity, attrs):
 
 # initialization hooks #########################################################
 
-class ServerMaintenanceHook(hook.Hook):
-    """install attribute map on the system source sql generator
+
+class VCSFileInitHook(hook.Hook):
+    """install attribute map on the system source sql generator and init
+    configuration
     """
-    __regid__ = 'vcsfile.servermaintenancehook'
-    events = ('server_maintenance',)
+    __regid__ = 'vcsfile.hook.init'
+    events = ('server_startup', 'server_maintenance')
+
     def __call__(self):
-        repo = self.repo
-        repo.system_source.set_storage('VersionContent', 'data',
+        self.repo.system_source.set_storage('VersionContent', 'data',
                                        bridge.VCSStorage())
-
-
-class ServerStartupHook(hook.Hook):
-    """synchronize repository on server startup, and install attribute map
-    on the system source sql generator
-    """
-    __regid__ = 'vcsfile.serverstartuphook'
-
-    events = ('server_startup',)
-    def __call__(self):
-        repo = self.repo
-        repo.system_source.set_storage('VersionContent', 'data',
-                                       bridge.VCSStorage())
-        cacheroot = repo_cache_dir(repo.config)
+        cacheroot = repo_cache_dir(self.repo.config)
         if not os.access(cacheroot, os.W_OK):
-            raise ValueError, ('directory %r is not writable '
-                               '(check  option)' % (cacheroot, CACHEROOT))
+            raise ValueError('directory %r is not writable (check  option)'
+                             % (cacheroot, CACHEROOT))
 
-        def refresh_local_hgrepo_caches(repo, osp=osp, bridge=bridge,
-                                        cacheroot=cacheroot):
+
+
+class VCSFileStartHook(hook.Hook):
+    """register task to regularly pull/import vcs repositories content"""
+    __regid__ = 'vcsfile.hook.start'
+    events = ('server_startup',)
+
+    def __call__(self):
+        repo = self.repo
+        def refresh_local_repo_caches(repo=repo, osp=osp, bridge=bridge):
             session = repo.internal_session()
             try:
                 reposrset = session.execute(
@@ -158,33 +148,25 @@ class ServerStartupHook(hook.Hook):
                 session.close()
 
         repo.looping_task(repo.config.get('check-revision-interval'),
-                          refresh_local_hgrepo_caches, repo)
+                          refresh_local_repo_caches)
 
 
 # internals to be able to create new vcs repo revision using rql queries #######
 
-class VCTransactionOp(hook.Operation):
+class VCTransactionOp(hook.LateOperation):
 
     def precommit_event(self):
+        self.transaction.precommit()
         bridge.set_at_revision(self.session, self.revision.eid)
-        transactions = self.session.transaction_data.setdefault('vctransactions', {})
-        for transaction in transactions.itervalues():
-            transaction.precommit()
 
     def revertprecommit_event(self):
-        transactions = self.session.transaction_data.pop('vctransactions', {})
-        for transaction in transactions.itervalues():
-            transaction.rollback()
+        self.transaction.rollback()
 
     def postcommit_event(self):
-        transactions = self.session.transaction_data.setdefault('vctransactions', {})
-        for transaction in transactions.itervalues():
-            transaction.commit()
+        self.transaction.commit()
 
     def rollback_event(self):
-        transactions = self.session.transaction_data.setdefault('vctransactions', {})
-        for transaction in transactions.itervalues():
-            transaction.rollback()
+        self.transaction.rollback()
 
 
 class VFModificationDateOp(hook.DataOperationMixIn, hook.Operation):
@@ -242,7 +224,7 @@ class AddRevisionHook(hook.Hook):
         transaction.reveid = entity.eid
         transactions[vcsrepoeid] = transaction
         entity.cw_edited['revision'] = transaction.rev
-        VCTransactionOp(session, revision=entity)
+        VCTransactionOp(session, revision=entity, transaction=transaction)
 
 
 class AddVersionContentHook(hook.Hook):
@@ -267,7 +249,7 @@ class AddVersionContentHook(hook.Hook):
 class AddDeletedVersionContentHook(hook.Hook):
     __regid__ = 'vcsfile.add_deleted_version_content_hook'
     __select__ = (hook.Hook.__select__ & is_instance('DeletedVersionContent')
-                  & is_user_session())
+                  & session.is_user_session())
     events = ('before_add_entity',)
 
     def __call__(self):
@@ -280,7 +262,7 @@ class AddDeletedVersionContentHook(hook.Hook):
 class AddVersionedFileHook(hook.Hook):
     __regid__ = 'vcsfile.add_versioned_file_hook'
     __select__ = (hook.Hook.__select__ & is_instance('VersionedFile')
-                  & is_user_session())
+                  & session.is_user_session())
     events = ('before_add_entity',)
 
     def __call__(self):
@@ -291,7 +273,7 @@ class SetRepoCacheBeforeAddHook(hook.Hook):
     """clone just after repo creation
     mostly helps the tests
     """
-    __regid__ = 'clone_hgrepo_after_add'
+    __regid__ = 'clone_repo_after_add'
     __select__ = (hook.Hook.__select__ & is_instance('Repository'))
     events = ('before_add_entity',)
     order = 2
@@ -306,7 +288,7 @@ class CloneRepoAfterAddHook(hook.Hook):
     """clone just after repo creation, to allow operation which needs local
     access and get quicker checkout if test are running on the same host
     """
-    __regid__ = 'clone_hgrepo_after_add'
+    __regid__ = 'clone_repo_after_add'
     __select__ = (hook.Hook.__select__ & is_instance('Repository'))
     events = ('after_add_entity',)
     order = 2
@@ -351,7 +333,7 @@ class DeleteDirsOp(hook.DataOperationMixIn, hook.Operation):
             try:
                 shutil.rmtree(cachedir)
                 self.info('deleted repository cache at %s', cachedir)
-            except:
+            except Exception:
                 self.exception('cant delete repository cache at %s', cachedir)
 
 class DeleteRepositoryHook(hook.Hook):
@@ -364,7 +346,7 @@ class DeleteRepositoryHook(hook.Hook):
     def __call__(self):
         if self.entity.local_cache:
             cachedir = osp.dirname(osp.join(
-                self._cw.vreg.config['local-repo-cache-root'],
+                self._cw.vreg.config.local_repo_cache_root,
                 self.entity.local_cache)).encode('ascii')
             DeleteDirsOp.get_instance(self._cw).add_data(cachedir)
 
@@ -412,7 +394,7 @@ class CheckImmutableAttributeHook(hook.Hook):
     __regid__ = 'vcsfile.check_immutable_attribute_hook'
     __select__ = (hook.Hook.__select__
                   & is_instance('Revision', 'DeletedVersionContent', 'VersionContent')
-                  & is_user_session())
+                  & session.is_user_session())
     events = ('before_update_entity',)
 
     def __call__(self):
@@ -443,7 +425,7 @@ class AddUpdateRepositoryHook(hook.Hook):
         if entity.source_url and not entity.path and not entity.local_cache:
             try:
                 url_to_relative_path(entity.source_url)
-            except:
+            except Exception:
                 msg = self._cw._('url must be of the form protocol://path/to/stuff')
                 raise ValidationError(entity.eid, {'source_url': msg})
 
