@@ -18,7 +18,7 @@ from mercurial.node import nullid, short as short_hex
 
 from logilab.common import tempattr
 
-from cubicweb import Binary, QueryError, ExecutionError
+from cubicweb import Binary, QueryError, ValidationError, ExecutionError
 from cubicweb.server.ssplanner import EditedEntity
 
 from cubes.vcsfile import repo, bridge
@@ -32,6 +32,11 @@ def monkeypatch_mercurial_url():
         del httphandler.__del__
 monkeypatch_mercurial_url()
 
+def spop(aset, item):
+    try:
+        aset.remove(item)
+    except KeyError:
+        pass
 
 class vcsfile_ui(ui.ui):
     """ redirect output to cubicweb log """
@@ -116,7 +121,7 @@ class HGRepository(repo.VCSRepository):
         for i in xrange(latestrev + 1, actualrev + 1):
             try:
                 self.import_revision(repoentity, repo, i)
-            except:
+            except Exception:
                 self.critical('error while importing revision %s of %s',
                               i, self.path, exc_info=True)
                 raise
@@ -146,18 +151,21 @@ class HGRepository(repo.VCSRepository):
                    'changeset': changeset,
                    'branch': unicode(ctx.branch()),
                    }
-        parents = [short_hex(n) for n in repo.changelog.parents(node)
+        parent_changesets = [short_hex(n) for n in repo.changelog.parents(node)
                    if n != nullid]
-        if parents:
-            if len(parents) == 1:
-                parents = execute('Revision X WHERE X changeset %(cs)s, '
-                                  'X from_repository R, R eid %(r)s',
-                                  {'cs': parents[0], 'r': self.eid})
-            else:
-                parents = execute('Revision X WHERE X changeset IN (%s), '
-                                  'X from_repository R, R eid %%(r)s'
-                                  % ','.join("'%s'"%cs for cs in parents),
-                                  {'r': self.eid})
+        if not parent_changesets:
+            parents = []
+        elif len(parent_changesets) == 1:
+            parents = execute(
+                'Any X,XC,XR WHERE X revision XR, X changeset XC, '
+                'X changeset %(cs)s, X from_repository R, R eid %(r)s',
+                {'cs': parent_changesets[0], 'r': self.eid})
+        else:
+            parents = execute(
+                'Any X,XC,XR WHERE X revision XR, X changeset XC, '
+                'X changeset IN (%s), X from_repository R, R eid %%(r)s'
+                % ','.join("'%s'"%cs for cs in parent_changesets),
+                {'r': self.eid})
         revdata['parents'] = [r[0] for r in parents]
         reveid = bridge.import_revision(session, self.eid, **revdata)
         if not repoentity.import_revision_content:
@@ -177,7 +185,9 @@ class HGRepository(repo.VCSRepository):
                 continue
             bridge.import_deleted_version_content(session, self.eid, reveid,
                                                   upath, date)
-        bridge.set_at_revision(session, reveid)
+        parent_idx = dict( ((cs, lrev) for _, cs, lrev in parents) )
+        parent_lrevs = [parent_idx[cs] for cs in parent_changesets]
+        bridge.set_at_revision(session, reveid, parent_lrevs)
 
     def _import_version_content(self, session, i, reveid, removed, path, upath,
                                 date, repo, ctx):
@@ -197,9 +207,9 @@ class HGRepository(repo.VCSRepository):
             else:
                 self.error('detected copy or rename of %s@%s but unable'
                            ' to find associated version content',
-                           oldfile, pcs)
-        bridge.import_version_content(session, self.eid, reveid, upath,
-                                      date, **vcdata)
+                           oldfile, rset.args['cs'])
+        return bridge.import_version_content(session, self.eid, reveid, upath,
+                                             date, **vcdata)
 
     def _renamed_vc_rset(self, session, repo, renameinfo):
         oldfile, fileid = renameinfo
@@ -245,7 +255,8 @@ class HGRepository(repo.VCSRepository):
 
     # reimport #################################################################
 
-    def check(self, repoentity, start_revision=0, commitevery=1):
+    def check(self, repoentity, start_revision=0, commitevery=1,
+              strip=True, import_new=False):
         """import content from the mercurial repository"""
         assert self.eid == repoentity.eid
         if not exists(self.path):
@@ -257,50 +268,106 @@ class HGRepository(repo.VCSRepository):
         except Exception, ex:
             self.exception('Impossible to open repository %s (%s)', self, ex)
             return
+        session = repoentity._cw
+        if strip and start_revision != 0:
+            raise Exception('must start from revision 0 to strip')
+        if strip and not repoentity.import_revision_content:
+            for etype in ('DeletedVersionContent', 'VersionContent', 'VersionedFile'):
+                rset = session.execute(
+                    'DELETE %s X WHERE X from_repository R, R eid %%(r)s' % etype,
+                    {'r': self.eid})
+                if rset:
+                    self.warning("deleted %s %s (content shouldn't be imported)",
+                                 len(rset), session._(etype))
+        else:
+            rset = session.execute(
+                'DELETE VersionedFile X WHERE X from_repository R, R eid %(r)s, '
+                'NOT VC content_for X', {'r': self.eid})
+            if rset:
+                self.warning('deleted %s files without any known revision',
+                             len(rset))
+        if strip and start_revision == 0: # full check
+            stripidx = {}
+            stripidx['Revision'] = set(
+                (eid for eid, in session.execute(
+                    'Revision X WHERE X from_repository R, R eid %(r)s',
+                    {'r': self.eid})) )
+            stripidx['VersionedFile'] = set(
+                (eid for eid, in session.execute(
+                    'VersionedFile X WHERE X from_repository R, R eid %(r)s',
+                    {'r': self.eid})) )
+        else:
+            stripidx = None
         # actual rev is the *number* of revision, latest rev the highest
         # revision number while hg revisions start at 0 -> - 1
         actualrev = bw_nb_revisions(repo.changelog) - 1
-        session = repoentity._cw
         latestrev = repoentity.latest_known_revision()
         self.info('hg repo %s: known revisions up to %s, repo revision %s',
                   self.path, latestrev, actualrev)
+        if not import_new:
+            actualrev = latestrev
         for i in xrange(start_revision, actualrev + 1):
             if i > latestrev:
                 self.import_revision(repoentity, repo, i)
             else:
-                self.check_revision(repoentity, repo, i)
+                self.check_revision(repoentity, repo, i, stripidx)
             if not i % commitevery:
                 # give a change to threads waiting for a pool by freeing it then
                 # reacquiring it
                 session.commit()
                 session.set_pool()
+        if strip:
+            if stripidx['Revision']:
+                self.warning('deleted %s revisions not in repository',
+                             len(stripidx['Revision']))
+                session.repo.glob_delete_entities(session, stripidx['Revision'])
+            if stripidx['VersionedFile']:
+                self.warning('deleted %s files not in repository',
+                             len(stripidx['VersionedFile']))
+                session.repo.glob_delete_entities(session, stripidx['VersionedFile'])
         session.commit()
 
-    def check_revision(self, repoentity, repo, i):
+    def check_revision(self, repoentity, repo, i, stripidx):
         self.info('checking revision %s', i)
         session = repoentity._cw
         execute = session.execute
         node = repo.changelog.node(i)
         changeset = unicode(short_hex(node))
-        rev = execute('Any X WHERE X from_repository R, R eid %(repo)s, X changeset %(cs)s',
-                      {'repo': repoentity.eid, 'cs': changeset})
-        if len(rev) > 1:
+        rset = execute('Any X,XR WHERE X revision XR, X from_repository R, '
+                       'R eid %(repo)s, X changeset %(cs)s',
+                       {'repo': repoentity.eid, 'cs': changeset})
+        if len(rset) > 1:
             # will crash on entities referencing this revision has parent,
             # better to stop now
-            raise ExecutionError('revision #%s has more than one related entities (%s). '
-                                 'Fix this manually first' % (
-                                     (changeset, ','.join(str(e) for e, in rev))))
-        if not rev:
-            rev = execute('Any X WHERE X from_repository R, R eid %(repo)s, X revision %(rev)s, X changeset ULL',
-                          {'repo': repoentity.eid, 'rev': i})
-            if rev:
+            raise ExecutionError('revision #%s has more than one related '
+                                 'entities (%s). Fix this manually first'
+                                 % (changeset, rset))
+        if rset:
+            rev = rset.get_entity(0, 0)
+            if rev.revision != i:
+                self.warning('revision for changeset #%s has local revision '
+                             'number to %s instead of %s',
+                             changeset, rev.revision, i)
+                # ensure we won't break (from repository, revision) unique
+                # together constraint
+                execute('SET X revision %(revneg)s WHERE X from_repository R, '
+                        'R eid %(repo)s, X revision %(rev)s',
+                        {'repo': repoentity.eid, 'rev': i, 'revneg': -i})
+                rev.set_attributes(revision=i)
+        else:
+            rset = execute('Any X WHERE X from_repository R, X changeset NULL, '
+                           'R eid %(repo)s, X revision %(rev)s',
+                           {'repo': repoentity.eid, 'rev': i})
+            if rset:
                 self.warning('revision missing changeset: %s (#%s)', i, changeset)
-                rev.get_entity(0, 0).set_attributes(changeset=changeset)
-        if not rev:
-            self.warning('revision %s isn\'t imported', changeset)
-            self.import_revision(repoentity, repo, i)
-            return
-        rev = rev.get_entity(0, 0)
+                rev = rset.get_entity(0, 0)
+                rev.set_attributes(changeset=changeset)
+            else:
+                self.warning('revision %s isn\'t imported', changeset)
+                self.import_revision(repoentity, repo, i)
+                return
+        if stripidx is not None:
+            stripidx['Revision'].remove(rev.eid)
         ctx = repo.changectx(i)
         date = datetime.fromtimestamp(ctx.date()[0])
         # XXX check revision's attributes
@@ -310,27 +377,37 @@ class HGRepository(repo.VCSRepository):
         #            'changeset': changeset,
         #            'branch': unicode(ctx.branch()),
         #            }
-        parents = [short_hex(n) for n in repo.changelog.parents(node)
-                   if n != nullid]
-        if parents:
-            if len(parents) == 1:
-                parentsrset = execute('Revision X WHERE X changeset %(cs)s, '
-                                  'X from_repository R, R eid %(r)s',
-                                  {'cs': parents[0], 'r': self.eid})
+        parent_changesets = [short_hex(n) for n in repo.changelog.parents(node)
+                             if n != nullid]
+        if parent_changesets:
+            args = {'r': self.eid}
+            if len(parent_changesets) == 1:
+                restr = 'X changeset %(cs)s'
+                args['cs'] = parent_changesets[0]
             else:
-                parentsrset = execute('Revision X WHERE X changeset IN (%s), '
-                                  'X from_repository R, R eid %%(r)s'
-                                  % ','.join("'%s'"%cs for cs in parents),
-                                  {'r': self.eid})
-            assert len(parentsrset) == len(parents), 'missing parent: expected %s, got %s' % (
-                parents, [e.changeset for e in parentsrset.entities()])
-            for preveid, in parentsrset:
+                restr = 'X changeset IN (%s)' % ','.join(
+                    repr(cs) for cs in parent_changesets)
+            parents = execute(
+                'Any X,XC,XR WHERE X changeset XC, X revision XR, '
+                'X from_repository R, R eid %%(r)s, %s' % restr, args)
+            assert len(parents) == len(parent_changesets), 'missing parent: expected %s, got %s' % (
+                parents, [e.changeset for e in parents.entities()])
+            for values in parents:
                 if execute('SET R parent_revision PR WHERE R eid %(r)s, PR eid %(pr)s, '
                            'NOT R parent_revision PR',
-                           {'r': rev.eid, 'pr': preveid}):
+                           {'r': rev.eid, 'pr': values[0]}):
                     self.warning('add missing parent revision')
+            args['r'] = rev.eid
+            if execute('DELETE R parent_revision X WHERE R parent_revision X, '
+                       'R eid %%(r)s, NOT %s' % restr, args):
+                self.warning('delete erroneous parent revision')
+        else:
+            parents = []
         if not repoentity.import_revision_content:
             return
+        contenteids = set(
+            (eid for eid, in session.execute(
+                'Any X WHERE X from_revision R, R eid %(r)s', {'r': rev.eid})) )
         changes = repo.status(ctx.parents()[0].node(), ctx.node())[:3]
         modified, added, removed = changes
         path_filter = repoentity.subpath
@@ -339,14 +416,19 @@ class HGRepository(repo.VCSRepository):
             if not (path_filter is None or upath.startswith(path_filter)):
                 continue
             directory, fname = split(path)
-            vc = execute('Any VC WHERE VF directory %(d)s, VF name %(n)s, '
+            vc = execute('Any VC,VF WHERE VF directory %(d)s, VF name %(n)s, '
                          'VC is VersionContent, VC content_for VF, VC from_revision R, R eid %(r)s',
                          {'d': directory, 'n': fname, 'r': rev.eid})
             if not vc:
                 self.warning('add missing file %s modified/added by revision', upath)
-                self._import_version_content(session, i, rev.eid, removed,
-                                             path, upath, date, repo, ctx)
+                vceid, vfeid = self._import_version_content(
+                    session, i, rev.eid, removed, path, upath, date, repo, ctx)
+                if stripidx is not None:
+                    spop(stripidx['VersionedFile'], vfeid)
             else:
+                if stripidx is not None:
+                    spop(stripidx['VersionedFile'], vc[0][1])
+                contenteids.remove(vc[0][0])
                 filectx = ctx[path]
                 renameinfo = filectx.renamed()
                 if renameinfo:
@@ -354,30 +436,45 @@ class HGRepository(repo.VCSRepository):
                     if pvc:
                         assert len(pvc) == 1
                         if oldfile in removed:
-                            if execute('SET X vc_rename Y WHERE X eid %(x)s, Y eid %(y)s, NOT X vc_rename Y',
+                            if execute('SET X vc_rename Y WHERE X eid %(x)s, '
+                                       'Y eid %(y)s, NOT X vc_rename Y',
                                        {'x': vc[0][0], 'y': pvc[0][0]}):
                                 self.warning('add rename information on %s', upath)
                         else:
-                            if execute('SET X vc_copy Y WHERE X eid %(x)s, Y eid %(y)s, NOT X vc_copy Y',
+                            if execute('SET X vc_copy Y WHERE X eid %(x)s, '
+                                       'Y eid %(y)s, NOT X vc_copy Y',
                                        {'x': vc[0][0], 'y': pvc[0][0]}):
                                 self.warning('add copy information on %s', upath)
                     else:
                         self.error('detected copy or rename of %s@%s but unable'
                                    ' to find associated version content',
-                                   oldfile, pcs)
+                                   oldfile, pvc.args['cs'])
         for path in removed:
             upath = unicode(path, self.encoding, 'replace')
             if not (path_filter is None or upath.startswith(path_filter)):
                 continue
             directory, fname = split(path)
-            vc = execute('Any VC WHERE VF directory %(d)s, VF name %(n)s, '
-                         'VC is DeletedVersionContent, VC content_for VF, VC from_revision R, R eid %(r)s',
+            vc = execute('Any VC,VF WHERE VF directory %(d)s, VF name %(n)s, '
+                         'VC is DeletedVersionContent, VC content_for VF, '
+                         'VC from_revision R, R eid %(r)s',
                          {'d': directory, 'n': fname, 'r': rev.eid})
             if not vc:
                 self.warning('add missing file %s deleted by revision', upath)
-                bridge.import_deleted_version_content(session, self.eid, rev.eid,
-                                                      upath, date)
-        bridge.set_at_revision(session, rev.eid, safetybelt=True)
+                vceid, vfeid = bridge.import_deleted_version_content(
+                    session, self.eid, rev.eid, upath, date)
+                if stripidx is not None:
+                    spop(stripidx['VersionedFile'], vfeid)
+            else:
+                if stripidx is not None:
+                    spop(stripidx['VersionedFile'], vc[0][1])
+                contenteids.remove(vc[0][0])
+        if contenteids:
+            self.warning('delete %s file revisions actually not modified by this revision',
+                         len(contenteids))
+            session.repo.glob_delete_entities(session, contenteids)
+        parent_idx = dict( ((cs, lrev) for _, cs, lrev in parents) )
+        parent_lrevs = [parent_idx[cs] for cs in parent_changesets]
+        bridge.set_at_revision(session, rev.eid, parent_lrevs, safetybelt=True)
 
 
 class HGTransaction(object):
@@ -413,6 +510,21 @@ class HGTransaction(object):
     def precommit(self):
         if not self.changes:
             raise QueryError('nothing changed')
+        # XXX to what commit does here + strip in revert_precommit
+        branch = self.repohdlr.encode(
+            self.revision.cw_attr_cache.get('branch') or u'default')
+        if branch != u'default':
+            self.extra['branch'] = branch
+        try:
+            self.hgparent = p1 = self.repo.branchtags()[branch]
+        except KeyError:
+            # new branch
+            self.hgparent = p1 = self.repo.branchtags().get('default')
+        if p1:
+            if not self.revision.parent_revision:
+                raise ValidationError(self.revision.eid, {'parent_revision': 'missing expected parent'})
+            if [short_hex(p1)] != [r.changeset for r in self.revision.parent_revision]:
+                raise ValidationError(None, {None: 'concurrency error, please re-upload the revision'})
 
     def commit(self):
         # XXX: make umask configurable
@@ -424,29 +536,22 @@ class HGTransaction(object):
 
     def _commit(self):
         # XXX merging branches
-        encode = self.repohdlr.encode
-        branch = encode(self.revision.cw_attr_cache.get('branch') or u'default')
-        if branch != u'default':
-            self.extra['branch'] = branch
-        author = encode(self.revision.cw_attr_cache.get('author', u''))
-        msg = encode(self.revision.cw_attr_cache.get('description', u''))
-        try:
-            p1 = self.repo.branchtags()[branch]
-        except KeyError:
-            # new branch
-            p1 = self.repo.branchtags().get('default')
-        p2 = None
+        repohdlr = self.repohdlr
+        revision = self.revision
+        author = repohdlr.encode(revision.cw_attr_cache.get('author', u''))
+        msg = repohdlr.encode(revision.cw_attr_cache.get('description', u''))
         # ensure mercurial will use configured repo encoding, not locale's
         # encoding
         # XXX modifying module's global is not very nice but I've no other idea
         try:
             # mercurial >= 1.3.1
             from mercurial import encoding
-            encoding.encoding = self.repohdlr.encoding
+            encoding.encoding = repohdlr.encoding
         except ImportError:
-            util._encoding = self.repohdlr.encoding
-        ctx = context.memctx(self.repo, (p1, p2), msg, self.changes.keys(),
-                             self._filectx, author, extra=self.extra)
+            util._encoding = repohdlr.encoding
+        ctx = context.memctx(self.repo, (self.hgparent, None), msg,
+                             self.changes.keys(), self._filectx, author,
+                             extra=self.extra)
         try:
             # mercurial < 1.3.1
             node = self.repo._commitctx(ctx, force=True, force_editor=False,
@@ -460,7 +565,7 @@ class HGTransaction(object):
         # / rql repository)
         source = self.session.repo.system_source
         # remove previous cached value to only save the changeset
-        with tempattr(self.revision, 'cw_edited', EditedEntity(self.revision)) as rev:
+        with tempattr(revision, 'cw_edited', EditedEntity(revision)) as rev:
             rev.cw_edited['changeset'] = unicode(short_hex(node))
             source.update_entity(self.session, rev)
         # we need to explicitly commit the connection to the database since this
